#
# Copyright (c) 2022 LAAS/CNRS
# All rights reserved.
#
# Redistribution  and  use  in  source  and binary  forms,  with  or  without
# modification, are permitted provided that the following conditions are met:
#
#   1. Redistributions of  source  code must retain the  above copyright
#      notice and this list of conditions.
#   2. Redistributions in binary form must reproduce the above copyright
#      notice and  this list of  conditions in the  documentation and/or
#      other materials provided with the distribution.
#
# THE SOFTWARE  IS PROVIDED "AS IS"  AND THE AUTHOR  DISCLAIMS ALL WARRANTIES
# WITH  REGARD   TO  THIS  SOFTWARE  INCLUDING  ALL   IMPLIED  WARRANTIES  OF
# MERCHANTABILITY AND  FITNESS.  IN NO EVENT  SHALL THE AUTHOR  BE LIABLE FOR
# ANY  SPECIAL, DIRECT,  INDIRECT, OR  CONSEQUENTIAL DAMAGES  OR  ANY DAMAGES
# WHATSOEVER  RESULTING FROM  LOSS OF  USE, DATA  OR PROFITS,  WHETHER  IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR  OTHER TORTIOUS ACTION, ARISING OUT OF OR
# IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
#
#                                           Anthony Mallet on Mon Jan 17 2022
#
"""Events and exceptions."""

import asyncio
import selectors


# === Status ===============================================================
#
class Status(int):
  """Status of pending requests."""

  __map = { None: 0, 'none': 0, 'sent': 1, 'done': 2, 'error': 3 }

  def __new__(self, key):
    try: return getattr(self, str(key))
    except:
      try: return super().__new__(self, self.__map[key])
      except: raise AttributeError(key) from None

  @classmethod
  def items(cls):
    return iter(cls.__map)

  def __repr__(self):
    return next(str(k) for k, v in self.__map.items() if self == v)

# define Status constants
for k in Status.items(): setattr(Status, str(k), Status(k))


# === GenoMError ===========================================================
#
class GenoMError(RuntimeError):
  """Raised when an exception occurs in a remote component."""

  def __init__(self, error):
    self.ex = error.get('ex', None)
    self.detail = error.get('detail', None)
    super().__init__(self.ex, self.detail)

  def __repr__(self):
    if self.detail is not None:
      return repr({'ex': self.ex, 'detail': self.detail})
    return repr({'ex': self.ex})

  def __str__(self):
    if self.detail is not None:
      return str(self.ex) + ': ' + str(self.detail)
    return str(self.ex)


# === Mux ==================================================================
#
class Mux:
  """Events multiplexer class."""

  # event selector
  __srcs = selectors.PollSelector()

  # read chunks of that size
  CHUNK_SIZE = 4096

  # --- watch --------------------------------------------------------------
  #
  def watch(self, sock, callback):
    """Add socket to the list of event sources and register callback as the
    event handler. If callback is None, remove any existing callback instead.
    """
    if callback is not None:
      self.__srcs.register(sock, selectors.EVENT_READ, callback)
    else:
      try:
        self.__srcs.unregister(sock)
      except ValueError: pass # this happens with multiple close


  # --- wait ---------------------------------------------------------------
  #
  def wait(self, timeout = None):
    """Block until one event becomes available, or a timeout expires.

    Return True if at least one event was processed, None otherwise.
    """
    e = self.__srcs.select(timeout)
    if not e: return None

    for key, event in e:
      try:
        # since requests callbacks may be processed as the result of key.data()
        # invocation, a recursive wait() may have already read available data
        # on this socket. So a BlockingIOError is possible and can be ignored.
        # Similarly, a OSError may happen if the server has sent us a closing
        # notice while nested in key.data().
        data = key.fileobj.recv(self.CHUNK_SIZE)
        key.data(data)
      except (BlockingIOError, OSError): pass

    return True


  # --- __await__ ----------------------------------------------------------
  #
  def __await__(self):
    """Asynchronous wait until one event becomes available.
    """
    # Process any already available data
    if self.wait(timeout=0): return True

    # Build an asyncio list of readers and return when the first completes
    loop = asyncio.get_event_loop()
    w = [self.__arecv(loop, key) for key in self.__srcs.get_map().values()]

    yield from next(asyncio.as_completed(w)).__await__()
    return True


  # --- arecv --------------------------------------------------------------
  #
  async def __arecv(self, loop, key):
    """Asynchronously receive data from SelectorKey and invoke associated
    callback.
    """
    key.data(await loop.sock_recv(key.fileobj, self.CHUNK_SIZE))


  # --- __new__ ------------------------------------------------------------
  #
  __instance = None
  def __new__(cls):
    """Implements a singleton for this class."""
    if cls.__instance is None:
      cls.__instance = super().__new__(cls)
    return cls.__instance


mux = Mux()

def update(*, timeout = 0):
  """Process any pending genomix event."""
  return mux.wait(timeout)
